#include "api/df_fingerprint.h"
#include "datafox_library_iv.h"

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintAppendRecord( int connectionId, int deviceAddress, int templateType, const unsigned char *buffer, int bufferSize, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i", __func__, connectionId, deviceAddress, templateType );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || templateType < 0 || templateType > 3
        || (templateType == 0 && bufferSize > 161) 
        || (templateType == 1 && bufferSize > 216)
        || (templateType == 2 && bufferSize > 561)
        || (templateType == 3 && bufferSize != (11 + 624))
        || buffer == nullptr || errorNumber == nullptr || bufferSize < 20 )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fingertemplate pruefen (Gleichen sich in den ersten 11 Byte)
    StructFingertemplateDIN *fp = (StructFingertemplateDIN *)buffer;
    if ( fp->pidLo != 0 || fp->fid > 10 || (templateType != 3 && (fp->quality == 0 || fp->count == 0)) )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, bufferSize, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, bufferSize, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fingertemplate und size ausgeben.
    // Daten fuer Log in Puffer uebergeben
    unsigned char ucTmp[(11 + 624) * 3 + 1];
    memset( ucTmp, 0, sizeof(ucTmp) );
    if ( channelBridge->logEnabled() )
    {
        int i;
        for ( i = 0; i < bufferSize; ++i )
        {
            sprintf( (char*)&ucTmp[i * 3], "%02X ", (unsigned char)buffer[i] );
        }

        ucTmp[i * 3] = 0;
    }
    else
    {
        sprintf( (char*)ucTmp, "0x%p", buffer );
    }

    channelBridge->logAppend( "\t%s\t%i", ucTmp, bufferSize );

    int iRes = channelBridge->fingerprintAppendRecord( deviceAddress, templateType, buffer, bufferSize );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT);
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_MEMORY_FULL );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATES_NONE );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_NEED_PASSWORD );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PASSWORD_LOCK );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }

        iRes = 0;
    } 
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( deviceAddress, -iRes );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if (iRes == 0)
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError(channelBridge->getLastErrorNumber( deviceAddress ));
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintGetRecord( int connectionId, int deviceAddress, int templateType, int pid, int fid, unsigned char *buffer, int *bufferSize, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i\t%i\t%i", __func__, connectionId, deviceAddress, templateType, pid, fid );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || templateType < 0 || templateType > 3
        || pid == 0 || fid < 1  || fid > 10 || buffer == nullptr || bufferSize == nullptr || errorNumber == nullptr || *bufferSize < 20
        || (templateType == 0 && *bufferSize < 161) 
        || (templateType == 1 && *bufferSize < 216)
        || (templateType == 2 && *bufferSize < 561)
        || (templateType == 3 && *bufferSize < (11 + 624)) )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, (bufferSize == nullptr) ? 0 : *bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, *bufferSize, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    int iRes = channelBridge->fingerprintGetRecord( deviceAddress, (unsigned char&)templateType, pid, fid, buffer, (int&)*bufferSize );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT);
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_MEMORY_FULL );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATES_NONE );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_NEED_PASSWORD );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PASSWORD_LOCK );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( deviceAddress, -iRes );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    // Fingertemplate und size ausgeben.
    // Daten fuer Log in Puffer uebergeben
    unsigned char ucTmp[(11 + 624) * 3 + 1];
    memset( ucTmp, 0, sizeof( ucTmp ) );
    if ( iRes == 1 && channelBridge->logEnabled() )
    {
        int i;
        for ( i = 0; i < *bufferSize; ++i )
        {
            sprintf((char*)&ucTmp[i * 3], "%02X ", (unsigned char)buffer[i]);
        }

        ucTmp[i * 3] = 0;
    }
    else
    {
        sprintf( (char*)ucTmp, "0x%p", buffer );
    }

    channelBridge->logAppend( "\t%s\t%i", ucTmp, *bufferSize );

    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintDeleteRecord( int connectionId, int deviceAddress, int pid, int fid, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i\t%i", __func__, connectionId, deviceAddress, pid, fid );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || (pid == 0 && fid != 0) || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    int iRes = channelBridge->fingerprintDeleteRecord( deviceAddress, pid, fid );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT);
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_MEMORY_FULL );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATES_NONE );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_NEED_PASSWORD );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PASSWORD_LOCK );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( deviceAddress, -iRes );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintList( int connectionId, int deviceAddress, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t0x%08X", __func__, connectionId, deviceAddress, flags );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || (flags != 0 && flags != 0x01) || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    unsigned int recordCount;
    int iRes = channelBridge->fingerprintList( deviceAddress, flags, recordCount );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
        // Anschliessende Ermittlung der Datensatzanzahl erwuenscht?
        if ( flags & 0x01 )
        {
            iRes = recordCount;
        }
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT);
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_MEMORY_FULL );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATES_NONE );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_NEED_PASSWORD );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PASSWORD_LOCK );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
         channelBridge->setLastErrorNumber( deviceAddress, -iRes );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintBackup( int connectionId, int deviceAddress, const char *filePath, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i", __func__, connectionId, deviceAddress );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || filePath == 0 || flags < 0 || flags > 1 || errorNumber == nullptr ) 
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    int iRes = channelBridge->fingerprintBackup( deviceAddress, filePath, flags );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT);
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_MEMORY_FULL );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATES_NONE );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_NEED_PASSWORD );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PASSWORD_LOCK );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( deviceAddress, -iRes );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintRestore( int connectionId, int deviceAddress, const char *filePath, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i", __func__, connectionId, deviceAddress );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || filePath == 0 || flags < 0 || flags > 1 || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    int iRes = channelBridge->fingerprintRestore( deviceAddress, filePath, flags );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT);
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_MEMORY_FULL );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATES_NONE );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_NEED_PASSWORD );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PASSWORD_LOCK );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( deviceAddress, -iRes );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}
