/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef DATAFOX_COMMUNICATION_MODULE_APPLICATION_H
#define DATAFOX_COMMUNICATION_MODULE_APPLICATION_H

// Basisklasse einbinden
#include "datafox/communication/module_system.h"

// MODULE_APPLICATION( 2 )

#define IV_MODULE_APPLICATION                    0x0002

// Fehler:
// 1 -> Modul unbekannt
// 2 -> Funktion unbekannt
// 3 -> Zu wenige Parameter
// 4 -> Zu viele Parameter

// Clear: Alle Datensaetze loeschen. CLEAR_RECORDS( 0 )
#define IV_FUNCTION_CLEAR_RECORDS                0x0000
// Set: Datensatz quittieren. SET_RECORD( 1 )
#define IV_FUNCTION_SET_RECORD                   0x0001
// Get: Datensatz hohlen. GET_RECORD( 2 )
#define IV_FUNCTION_GET_RECORD                   0x0002

// Datensatzanfrage
//#define MB_CMD_READ_RECORD              0x40  /* Datensatz senden */

#define MB_ANSW_NO_RECORD               0xC0  /* kein Datensatz vorhanden */
#define MB_ANSW_OFFLINE_RECORD          0xC1  /* Offline Datensatz */
#define MB_ANSW_OLD_OFFLINE_RECORD      0xC3  /* wiederholt gesendeter Offline Datensatz */
#define MB_ANSW_ONLINE_RECORD           0xE1  /* Online Datensatz */
#define MB_ANSW_OLD_ONLINE_RECORD       0xE3  /* wiederholt gesendeter Online Datensatz */

// Datensatzquittierung
//#define MB_CMD_ACK_RECORD               0x48  /* Quittung fuer Datensatz */

#define MB_ANSW_ACK_OFF_RECORD	        0xC8  /* Quittung erhalten (fuer Offline-Datensatz) */
#define MB_ANSW_OLD_ACK_OFF_RECORD      0xCA  /* wiederholt gesendete Quittung */
#define MB_ANSW_ACK_ON_RECORD           0xE8  /* Quittung erhalten (fuer Online-Datensatz) */
#define MB_ANSW_OLD_ACK_ON_RECORD       0xEA  /* wiederholt gesendete Quittung */
#define MB_ANSW_TIMEOUT_ACK_RECORD      0xE9  /* Quittung kam nach Timeout (fuer Online-Datensatz) */
#define MB_ANSW_OLD_TIMEOUT_ACK_RECORD  0xEB  /* wiederholte Quittung nach Timeout */

// Datensatzloeschbefehl
//#define MB_CMD_DEL						0x49  /* Loescht alle Datensaetze. */



// Set: Message SHOW_MESSAGE( 3 )
#define IV_FUNCTION_SHOW_MESSAGE                 0x0003
// (UIN8(Time)unsigned char(Action)unsigned char(Audio)Message\0)
// -> ()
// Time ist die anzuzeigende Dauer
// Action ist die durchzufuehrende Aktion (unbenutzt und immer 0)
// Audio ist das zu verwendende akustische Signal
// Message ist der eigentliche Nachrichtentext

// Fehler:
// 1 -> Nachrichtentext zu lang fuer die Anzeige.
// 2 -> Es wird bereits eine Nachricht angezeigt.


// Set: Variable (Setup) SET_VARIABLE( 4 )
#define IV_FUNCTION_SET_VARIABLE                 0x0004
// Get: Variable (Setup) GET_VARIABLE( 5 )
#define IV_FUNCTION_GET_VARIABLE                 0x0005

// Der Name einer Variablen setzt sich auf folgende Art zusammen:
// MAJOR_NAME.MINOR_NAME
// Der Majorname einer Setupvariablen ist immer SETUP.
// Fehlernummer:
// 1 = MAJOR_NAME ist unbekannt
// 2 = MINOR_NAME ist unbekannt
// 3 = Länge des Variablenwertes ist zu lang.

#define IV_FUNCTION_SHOW_INFO                    0x0006
// Maschinennummer (1 Byte)
// Time ist die Anzeigedauer. (1 Byte) (0 im Hindergrund uebernehmen, -1 staendig anzeigen)
// Schriftgroessennummer (1 Byte)
// Akkustisches Signal (1 Byte)
// 16bit reserve (2 Byte)
// Flags z.b. 0x0001 (2 Byte) fuer Antiproportionale Schrift
// Infotext

#define IV_FUNCTION_ENTRANCE2_IDENTIFICATION     0x0007
// Alle Parameter mit fester Laenge. Linksausrichtung.
// param1: 3 Zeichen Tuermodul
// param2: 20 Zeichen Ausweis
// param3: 8 Zeichen Pin

// Aktive Antwort: Status der Zutrittspruefung als Parameter
#define IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK  0x0008

// Benachrichtigung ueber ein Ereignis
#define IV_FUNCTION_NOTIFY                       0x0009

// Mindestens ein Datensatz liegt vor.
#define IV_VALUE_NOTIFY_RECORD                   0x00

class CNotifyRecord 
{
public:
    CNotifyRecord() {
        channelID = 0;
        deviceID = 0;
        type = 0;
        priority = 0;
        memset(timestamp, 0, sizeof(timestamp));
        dt = 0;
        erase = 0;
    }

public:
    int channelID;      // Kanalnummer des Geraetes
    int deviceID;       // Busnummer des Geraetes
    unsigned char type; // Meldungstyp
    short priority;     // Prioritaet des Geraetes
    char timestamp[24]; // Zeitstempel der Meldungserzeugung (Format: "YYYY-MM-DD hh:mm:ss:mis")
    time_t dt;          // Zeitstempel der Annahme
    int erase;          // Wie der Dateninhalt zu handhaben ist. 0 -> Hinzufuegen, 1 -> Entfernen

public:
    bool operator < (const CNotifyRecord &other) const
    {
        return (priority < other.priority);
    }
};

class CRecord
{
public:
    CRecord() 
    {
        m_priority = 0;
        m_status = 0;
        memset(&m_timestamp, 0, sizeof(tm));
        m_delay = 0;
        m_length = 0;
        memset(m_data, 0, sizeof(m_data));
    }
    int m_priority;
    int m_status;
    struct tm m_timestamp;
    int m_delay;
    int m_length;
    unsigned char m_data[256];
};


#define IV_FUNCTION_ENTRANCE2_ONLINE_ACTION      0x000A
// Alle Parameter mit fester Laenge. Linksausrichtung.
// param1: 3 Zeichen Tuermodul
// param2: 4Byte Bitmaske: Bit0: Buzzer, Bit1: LED Grün, Bit2: LED Rot, Bit3: Relais 1, Bit4: Relais 2
// param3: 1Byte Typ. 0=Dauer Aus, 1=Dauer Ein, 2=Blinken (600ms an/aus)
// param4: 1Byte Dauer. Bei Typ = 2 ist 0=Dauernd ein oder 1 - 40 Sekunden (maximal 48Sekunden da 255*200ms)

// Restore: Datensatzzeiger auf ersten gueltigen Datensatz zurruecksetzen.
#define IV_FUNCTION_RESTORE_RECORDS              0x000B

// Den Type des zu verwendenden Zeichensatzes fuer Messages und Infos festsetzen
#define IV_FUNCTION_SET_TYPE_FONT                0x000C

// Anzahl der vorliegenden Datensaetze ermitteln.
#define IV_FUNCTION_RECORD_VOLUME                 0x000D

#define IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION     0x000E
// Alle Parameter mit fester Laenge. Linksausrichtung.
// param1: 2Byte MasterId
// param1: 3 Zeichen Tuermodul
// param2: 20 Zeichen Ausweis
// param3: 8 Zeichen Pin

// Aktive Antwort: Status der Zutrittspruefung als Parameter
#define IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK  0x000F

#define IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION      0x0010
// Alle Parameter mit fester Laenge. Linksausrichtung.
// param1: 2Byte MasterId
// param2: 3 Zeichen Tuermodul
// param3: 4Byte Bitmaske: Bit0: Buzzer, Bit1: LED Grün, Bit2: LED Rot, Bit3: Relais 1, Bit4: Relais 2
// param4: 1Byte Typ. 0=Dauer Aus, 1=Dauer Ein, 2=Blinken (600ms an/aus)
// param5: 1Byte Dauer. Bei Typ = 2 ist 0=Dauernd ein oder 1 - 40 Sekunden (maximal 48Sekunden da 255*200ms)

// In der Zutrittskontrolle Schließzylinder einlernen/löschen/etc..
#define IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND       0x0011
#define IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK    0x0012

// Kommandos für IV_FUNCTION_ACCESS_CONTROL_KNOB
#define IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_TEACH                   0
#define IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_DELETE                  1
#define IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST                    2
#define IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_CHANGE_BATTERY_BEGIN    3
#define IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_CHANGE_BATTERY_END      4

// Kundenspezifisches Komando ausführen und bestätigen
#define IV_FUNCTION_CUSTOM_COMMAND       0x0013
#define IV_FUNCTION_CUSTOM_COMMAND_OK    0x0014

// Kommandos Für Benutzerdefinierten Befehl
//#define IV_VALUE_COMMAND_WRITE                  0

// Kontainerklassen fuer Prioritaetswarteschlange importieren.
#include <deque>
#include <queue>

typedef std::deque<CRecord> record_deque;
typedef std::queue<CRecord> record_queue;

namespace Datafox
{
    namespace Communication
    {
        class Module_application : public Module_system
        {
        public:
            /// Konstruktor fuer Initialisierung der Klasse.
            Module_application( int connection_id = Connection_base::dv_connection_id, uint8_t device_address = Protocol_base::dv_device_address );

            /// Destruktor
            virtual ~Module_application();

        public:
            // Setzen der CallBack-Funktion fuer Datensatzbenachrichtigung
            void set_call_back_notify_record_available(int (*CBNotifyRecord)( CNotifyRecord & ))
            {
                m_call_back_notify_record_available = CBNotifyRecord;
            }

        protected:
            int (*m_call_back_notify_record_available)( CNotifyRecord & );


        public:
            // Gibt Modulfehler zurueck.
            virtual void error_module( uint16_t module, uint16_t error, std::string &text );
            // Gibt Funktionsfehlertext zurueck.
            virtual void error_function( uint16_t module, uint16_t func, uint16_t error, std::string &text );

            // MODULE_APPLICATION
            record_queue m_records;

            // Datensaetze loeschen
            int clear_records( unsigned char device_address );

            // Datensatzzeiger ruecksetzen
            int restore_records( unsigned char device_address );

            // Datensatz quittieren/anfordern
            int quit_record( uint8_t device_address, uint8_t &status );
            int read_record( uint8_t device_address, uint8_t *buffer, std::size_t &count, uint8_t &status );

            // Displaynachicht senden
            int write_message_to_display( unsigned char device_address, unsigned char time, unsigned char action, unsigned char audio, const std::string &text );

            // Variable setzen/lesen
            virtual int write_variable( unsigned char device_address, const std::string &name, const std::string &val );
            virtual int read_variable( unsigned char device_address, const std::string &name, std::string &val );

            // Info senden
            int write_memo_to_display( unsigned char device_address, unsigned char maschine, unsigned char time, unsigned char font, unsigned char audio, short flags, const std::string &text );

            // Zutrittsbuchung durchfuehren
            int access_control_identification( unsigned char device_address, const char *TM, const char *Id, const char *Pin, char *status, int &status_length );
            int reply_access_control_identification( unsigned char device_address );
            // LED und Relais auf Tuermodul schalten
            int access_control_online_action( unsigned char device_address, const char *TM, unsigned int mask, unsigned char type, unsigned char duration );

            // Eine Benachrichtigung melden
            int notify( unsigned char device_address, unsigned char type, int priority, char *timestamp, char *params );
            int notify_record_available( unsigned char device_address, int priority );

            // Zu verwendender Zeichensatz bei Anzeige von Message oder Info
            int write_font_type( unsigned char device_address, unsigned char type );

            // Anzahl der Datensaetze oder verwendetes Speichervolumen.
            int read_records_volume( unsigned char device_address,  int &number, int &memory_usage, int &memory_size );

            // Zutrittsbuchung durchfuehren
            int access_control_identification_ext( unsigned char device_address, int master_id, const char *TM, const char *Id, const char *Pin, char *status, int &status_length );
            int reply_access_control_identification_ext( unsigned char device_address );
            // LED und Relais auf Tuermodul schalten
            int access_control_online_action_ext( unsigned char device_address, int master_id, const char *TM, unsigned int mask, unsigned char type, unsigned char duration );

            // Zutrittskontroll-Befehl für Schließzylinder
            int access_control_knob_command( unsigned char device_address, unsigned char knob_command, unsigned short master_id, unsigned char knob_id, const char *params, unsigned char *data, int *size_data );
            int reply_access_control_knob_command( unsigned char device_address );

            // Kundenspezifischer Befehl
            int custom_command( unsigned char device_address, unsigned int identification, unsigned int custom_command, const char *data_in, int size_data_in, char *buffer_out, int *buffer_size_out );
            int reply_custom_command( unsigned char device_address );

            //#################################
            // CallRoutinen

            // Ausfuehren von Befehlen
            virtual int remote_call( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count );

        };

    } // namespace Communication
} // namespace Datafox

#endif
